var jsutils_inspect = require("../jsutils/inspect.js");

var jsutils_instanceOf = require("../jsutils/instanceOf.js");

var jsutils_toObjMap = require("../jsutils/toObjMap.js");

var language_directiveLocation = require("../language/directiveLocation.js");

var type_assertName = require("./assertName.js");

var type_definition = require("./definition.js");

var type_scalars = require("./scalars.js");

function isDirective(directive) {
  return jsutils_instanceOf.instanceOf(directive, GraphQLDirective);
}

class GraphQLDirective {
  constructor(config) {
    var _config$isRepeatable, _config$args;
    this.name = type_assertName.assertName(config.name);
    this.description = config.description;
    this.locations = config.locations;
    this.isRepeatable = null !== (_config$isRepeatable = config.isRepeatable) && void 0 !== _config$isRepeatable ? _config$isRepeatable : !1;
    this.extensions = jsutils_toObjMap.toObjMap(config.extensions);
    this.astNode = config.astNode;
    const args = null !== (_config$args = config.args) && void 0 !== _config$args ? _config$args : {};
    this.args = type_definition.defineArguments(args);
  }
  toConfig() {
    return {
      name: this.name,
      description: this.description,
      locations: this.locations,
      args: type_definition.argsToArgsConfig(this.args),
      isRepeatable: this.isRepeatable,
      extensions: this.extensions,
      astNode: this.astNode
    };
  }
  toString() {
    return "@" + this.name;
  }
  toJSON() {
    return this.toString();
  }
}

const GraphQLIncludeDirective = new GraphQLDirective({
  name: "include",
  description: "Directs the executor to include this field or fragment only when the `if` argument is true.",
  locations: [ language_directiveLocation.DirectiveLocation.FIELD, language_directiveLocation.DirectiveLocation.FRAGMENT_SPREAD, language_directiveLocation.DirectiveLocation.INLINE_FRAGMENT ],
  args: {
    if: {
      type: new type_definition.GraphQLNonNull(type_scalars.GraphQLBoolean),
      description: "Included when true."
    }
  }
});

const GraphQLSkipDirective = new GraphQLDirective({
  name: "skip",
  description: "Directs the executor to skip this field or fragment when the `if` argument is true.",
  locations: [ language_directiveLocation.DirectiveLocation.FIELD, language_directiveLocation.DirectiveLocation.FRAGMENT_SPREAD, language_directiveLocation.DirectiveLocation.INLINE_FRAGMENT ],
  args: {
    if: {
      type: new type_definition.GraphQLNonNull(type_scalars.GraphQLBoolean),
      description: "Skipped when true."
    }
  }
});

const GraphQLDeprecatedDirective = new GraphQLDirective({
  name: "deprecated",
  description: "Marks an element of a GraphQL schema as no longer supported.",
  locations: [ language_directiveLocation.DirectiveLocation.FIELD_DEFINITION, language_directiveLocation.DirectiveLocation.ARGUMENT_DEFINITION, language_directiveLocation.DirectiveLocation.INPUT_FIELD_DEFINITION, language_directiveLocation.DirectiveLocation.ENUM_VALUE ],
  args: {
    reason: {
      type: type_scalars.GraphQLString,
      description: "Explains why this element was deprecated, usually also including a suggestion for how to access supported similar data. Formatted using the Markdown syntax, as specified by [CommonMark](https://commonmark.org/).",
      defaultValue: "No longer supported"
    }
  }
});

const GraphQLSpecifiedByDirective = new GraphQLDirective({
  name: "specifiedBy",
  description: "Exposes a URL that specifies the behavior of this scalar.",
  locations: [ language_directiveLocation.DirectiveLocation.SCALAR ],
  args: {
    url: {
      type: new type_definition.GraphQLNonNull(type_scalars.GraphQLString),
      description: "The URL that specifies the behavior of this scalar."
    }
  }
});

const specifiedDirectives = [ GraphQLIncludeDirective, GraphQLSkipDirective, GraphQLDeprecatedDirective, GraphQLSpecifiedByDirective ];

exports.DEFAULT_DEPRECATION_REASON = "No longer supported";

exports.GraphQLDeprecatedDirective = GraphQLDeprecatedDirective;

exports.GraphQLDirective = GraphQLDirective;

exports.GraphQLIncludeDirective = GraphQLIncludeDirective;

exports.GraphQLSkipDirective = GraphQLSkipDirective;

exports.GraphQLSpecifiedByDirective = GraphQLSpecifiedByDirective;

exports.assertDirective = function assertDirective(directive) {
  if (!isDirective(directive)) {
    throw new Error(`Expected ${jsutils_inspect.inspect(directive)} to be a GraphQL directive.`);
  }
  return directive;
};

exports.isDirective = isDirective;

exports.isSpecifiedDirective = function isSpecifiedDirective(directive) {
  return specifiedDirectives.some((({name}) => name === directive.name));
};

exports.specifiedDirectives = specifiedDirectives;
