import { inspect } from "../jsutils/inspect.mjs";

import { invariant } from "../jsutils/invariant.mjs";

import { isIterableObject } from "../jsutils/isIterableObject.mjs";

import { isObjectLike } from "../jsutils/isObjectLike.mjs";

import { isNonNullType, isListType, isInputObjectType, isLeafType, isEnumType } from "../type/definition.mjs";

import { GraphQLID } from "../type/scalars.mjs";

import { e } from "../shared.mjs";

function astFromValue(value, type) {
  if (isNonNullType(type)) {
    const astValue = astFromValue(value, type.ofType);
    if ((null == astValue ? void 0 : astValue.kind) === e.NULL) {
      return null;
    }
    return astValue;
  }
  if (null === value) {
    return {
      kind: e.NULL
    };
  }
  if (void 0 === value) {
    return null;
  }
  if (isListType(type)) {
    const itemType = type.ofType;
    if (isIterableObject(value)) {
      const valuesNodes = [];
      for (const item of value) {
        const itemNode = astFromValue(item, itemType);
        if (null != itemNode) {
          valuesNodes.push(itemNode);
        }
      }
      return {
        kind: e.LIST,
        values: valuesNodes
      };
    }
    return astFromValue(value, itemType);
  }
  if (isInputObjectType(type)) {
    if (!isObjectLike(value)) {
      return null;
    }
    const fieldNodes = [];
    for (const field of Object.values(type.getFields())) {
      const fieldValue = astFromValue(value[field.name], field.type);
      if (fieldValue) {
        fieldNodes.push({
          kind: e.OBJECT_FIELD,
          name: {
            kind: e.NAME,
            value: field.name
          },
          value: fieldValue
        });
      }
    }
    return {
      kind: e.OBJECT,
      fields: fieldNodes
    };
  }
  if (isLeafType(type)) {
    const serialized = type.serialize(value);
    if (null == serialized) {
      return null;
    }
    if ("boolean" == typeof serialized) {
      return {
        kind: e.BOOLEAN,
        value: serialized
      };
    }
    if ("number" == typeof serialized && Number.isFinite(serialized)) {
      const stringNum = String(serialized);
      return integerStringRegExp.test(stringNum) ? {
        kind: e.INT,
        value: stringNum
      } : {
        kind: e.FLOAT,
        value: stringNum
      };
    }
    if ("string" == typeof serialized) {
      if (isEnumType(type)) {
        return {
          kind: e.ENUM,
          value: serialized
        };
      }
      if (type === GraphQLID && integerStringRegExp.test(serialized)) {
        return {
          kind: e.INT,
          value: serialized
        };
      }
      return {
        kind: e.STRING,
        value: serialized
      };
    }
    throw new TypeError(`Cannot convert value to AST: ${inspect(serialized)}.`);
  }
  invariant(!1);
}

const integerStringRegExp = /^-?(?:0|[1-9][0-9]*)$/;

export { astFromValue };
