var jsutils_inspect = require("../jsutils/inspect.js");

var jsutils_keyValMap = require("../jsutils/keyValMap.js");

var type_definition = require("../type/definition.js");

var type_directives = require("../type/directives.js");

var type_introspection = require("../type/introspection.js");

var type_scalars = require("../type/scalars.js");

var type_schema = require("../type/schema.js");

var utilities_valueFromAST = require("./valueFromAST.js");

var shared = require("../shared.js");

exports.buildClientSchema = function buildClientSchema(introspection, options) {
  const schemaIntrospection = introspection.__schema;
  const typeMap = jsutils_keyValMap.keyValMap(schemaIntrospection.types, (typeIntrospection => typeIntrospection.name), (typeIntrospection => function buildType(type) {
    if (null != type && null != type.name && null != type.kind) {
      switch (type.kind) {
       case type_introspection.TypeKind.SCALAR:
        return function buildScalarDef(scalarIntrospection) {
          return new type_definition.GraphQLScalarType({
            name: scalarIntrospection.name,
            description: scalarIntrospection.description,
            specifiedByURL: scalarIntrospection.specifiedByURL
          });
        }(type);

       case type_introspection.TypeKind.OBJECT:
        return function buildObjectDef(objectIntrospection) {
          return new type_definition.GraphQLObjectType({
            name: objectIntrospection.name,
            description: objectIntrospection.description,
            interfaces: () => buildImplementationsList(objectIntrospection),
            fields: () => buildFieldDefMap(objectIntrospection)
          });
        }(type);

       case type_introspection.TypeKind.INTERFACE:
        return function buildInterfaceDef(interfaceIntrospection) {
          return new type_definition.GraphQLInterfaceType({
            name: interfaceIntrospection.name,
            description: interfaceIntrospection.description,
            interfaces: () => buildImplementationsList(interfaceIntrospection),
            fields: () => buildFieldDefMap(interfaceIntrospection)
          });
        }(type);

       case type_introspection.TypeKind.UNION:
        return function buildUnionDef(unionIntrospection) {
          if (!unionIntrospection.possibleTypes) {
            const unionIntrospectionStr = jsutils_inspect.inspect(unionIntrospection);
            throw new Error(`Introspection result missing possibleTypes: ${unionIntrospectionStr}.`);
          }
          return new type_definition.GraphQLUnionType({
            name: unionIntrospection.name,
            description: unionIntrospection.description,
            types: () => unionIntrospection.possibleTypes.map(getObjectType)
          });
        }(type);

       case type_introspection.TypeKind.ENUM:
        return function buildEnumDef(enumIntrospection) {
          if (!enumIntrospection.enumValues) {
            const enumIntrospectionStr = jsutils_inspect.inspect(enumIntrospection);
            throw new Error(`Introspection result missing enumValues: ${enumIntrospectionStr}.`);
          }
          return new type_definition.GraphQLEnumType({
            name: enumIntrospection.name,
            description: enumIntrospection.description,
            values: jsutils_keyValMap.keyValMap(enumIntrospection.enumValues, (valueIntrospection => valueIntrospection.name), (valueIntrospection => ({
              description: valueIntrospection.description,
              deprecationReason: valueIntrospection.deprecationReason
            })))
          });
        }(type);

       case type_introspection.TypeKind.INPUT_OBJECT:
        return function buildInputObjectDef(inputObjectIntrospection) {
          if (!inputObjectIntrospection.inputFields) {
            const inputObjectIntrospectionStr = jsutils_inspect.inspect(inputObjectIntrospection);
            throw new Error(`Introspection result missing inputFields: ${inputObjectIntrospectionStr}.`);
          }
          return new type_definition.GraphQLInputObjectType({
            name: inputObjectIntrospection.name,
            description: inputObjectIntrospection.description,
            fields: () => buildInputValueDefMap(inputObjectIntrospection.inputFields)
          });
        }(type);
      }
    }
    const typeStr = jsutils_inspect.inspect(type);
    throw new Error(`Invalid or incomplete introspection result. Ensure that a full introspection query is used in order to build a client schema: ${typeStr}.`);
  }(typeIntrospection)));
  for (const stdType of [ ...type_scalars.specifiedScalarTypes, ...type_introspection.introspectionTypes ]) {
    if (typeMap[stdType.name]) {
      typeMap[stdType.name] = stdType;
    }
  }
  const queryType = schemaIntrospection.queryType ? getObjectType(schemaIntrospection.queryType) : null;
  const mutationType = schemaIntrospection.mutationType ? getObjectType(schemaIntrospection.mutationType) : null;
  const subscriptionType = schemaIntrospection.subscriptionType ? getObjectType(schemaIntrospection.subscriptionType) : null;
  const directives = schemaIntrospection.directives ? schemaIntrospection.directives.map((function buildDirective(directiveIntrospection) {
    if (!directiveIntrospection.args) {
      const directiveIntrospectionStr = jsutils_inspect.inspect(directiveIntrospection);
      throw new Error(`Introspection result missing directive args: ${directiveIntrospectionStr}.`);
    }
    if (!directiveIntrospection.locations) {
      const directiveIntrospectionStr = jsutils_inspect.inspect(directiveIntrospection);
      throw new Error(`Introspection result missing directive locations: ${directiveIntrospectionStr}.`);
    }
    return new type_directives.GraphQLDirective({
      name: directiveIntrospection.name,
      description: directiveIntrospection.description,
      isRepeatable: directiveIntrospection.isRepeatable,
      locations: directiveIntrospection.locations.slice(),
      args: buildInputValueDefMap(directiveIntrospection.args)
    });
  })) : [];
  return new type_schema.GraphQLSchema({
    description: schemaIntrospection.description,
    query: queryType,
    mutation: mutationType,
    subscription: subscriptionType,
    types: Object.values(typeMap),
    directives,
    assumeValid: null == options ? void 0 : options.assumeValid
  });
  function getType(typeRef) {
    if (typeRef.kind === type_introspection.TypeKind.LIST) {
      const itemRef = typeRef.ofType;
      if (!itemRef) {
        throw new Error("Decorated type deeper than introspection query.");
      }
      return new type_definition.GraphQLList(getType(itemRef));
    }
    if (typeRef.kind === type_introspection.TypeKind.NON_NULL) {
      const nullableRef = typeRef.ofType;
      if (!nullableRef) {
        throw new Error("Decorated type deeper than introspection query.");
      }
      const nullableType = getType(nullableRef);
      return new type_definition.GraphQLNonNull(type_definition.assertNullableType(nullableType));
    }
    return getNamedType(typeRef);
  }
  function getNamedType(typeRef) {
    const typeName = typeRef.name;
    if (!typeName) {
      throw new Error(`Unknown type reference: ${jsutils_inspect.inspect(typeRef)}.`);
    }
    const type = typeMap[typeName];
    if (!type) {
      throw new Error(`Invalid or incomplete schema, unknown type: ${typeName}. Ensure that a full introspection query is used in order to build a client schema.`);
    }
    return type;
  }
  function getObjectType(typeRef) {
    return type_definition.assertObjectType(getNamedType(typeRef));
  }
  function getInterfaceType(typeRef) {
    return type_definition.assertInterfaceType(getNamedType(typeRef));
  }
  function buildImplementationsList(implementingIntrospection) {
    if (null === implementingIntrospection.interfaces && implementingIntrospection.kind === type_introspection.TypeKind.INTERFACE) {
      return [];
    }
    if (!implementingIntrospection.interfaces) {
      const implementingIntrospectionStr = jsutils_inspect.inspect(implementingIntrospection);
      throw new Error(`Introspection result missing interfaces: ${implementingIntrospectionStr}.`);
    }
    return implementingIntrospection.interfaces.map(getInterfaceType);
  }
  function buildFieldDefMap(typeIntrospection) {
    if (!typeIntrospection.fields) {
      throw new Error(`Introspection result missing fields: ${jsutils_inspect.inspect(typeIntrospection)}.`);
    }
    return jsutils_keyValMap.keyValMap(typeIntrospection.fields, (fieldIntrospection => fieldIntrospection.name), buildField);
  }
  function buildField(fieldIntrospection) {
    const type = getType(fieldIntrospection.type);
    if (!type_definition.isOutputType(type)) {
      const typeStr = jsutils_inspect.inspect(type);
      throw new Error(`Introspection must provide output type for fields, but received: ${typeStr}.`);
    }
    if (!fieldIntrospection.args) {
      const fieldIntrospectionStr = jsutils_inspect.inspect(fieldIntrospection);
      throw new Error(`Introspection result missing field args: ${fieldIntrospectionStr}.`);
    }
    return {
      description: fieldIntrospection.description,
      deprecationReason: fieldIntrospection.deprecationReason,
      type,
      args: buildInputValueDefMap(fieldIntrospection.args)
    };
  }
  function buildInputValueDefMap(inputValueIntrospections) {
    return jsutils_keyValMap.keyValMap(inputValueIntrospections, (inputValue => inputValue.name), buildInputValue);
  }
  function buildInputValue(inputValueIntrospection) {
    const type = getType(inputValueIntrospection.type);
    if (!type_definition.isInputType(type)) {
      const typeStr = jsutils_inspect.inspect(type);
      throw new Error(`Introspection must provide input type for arguments, but received: ${typeStr}.`);
    }
    const defaultValue = null != inputValueIntrospection.defaultValue ? utilities_valueFromAST.valueFromAST(shared.parseValue(inputValueIntrospection.defaultValue), type) : void 0;
    return {
      description: inputValueIntrospection.description,
      type,
      defaultValue,
      deprecationReason: inputValueIntrospection.deprecationReason
    };
  }
};
