import { invariant } from "../jsutils/invariant.mjs";

import { keyMap } from "../jsutils/keyMap.mjs";

import { isNonNullType, isListType, isInputObjectType, isLeafType } from "../type/definition.mjs";

import { e } from "../shared.mjs";

function valueFromAST(valueNode, type, variables) {
  if (!valueNode) {
    return;
  }
  if (valueNode.kind === e.VARIABLE) {
    const variableName = valueNode.name.value;
    if (null == variables || void 0 === variables[variableName]) {
      return;
    }
    const variableValue = variables[variableName];
    if (null === variableValue && isNonNullType(type)) {
      return;
    }
    return variableValue;
  }
  if (isNonNullType(type)) {
    if (valueNode.kind === e.NULL) {
      return;
    }
    return valueFromAST(valueNode, type.ofType, variables);
  }
  if (valueNode.kind === e.NULL) {
    return null;
  }
  if (isListType(type)) {
    const itemType = type.ofType;
    if (valueNode.kind === e.LIST) {
      const coercedValues = [];
      for (const itemNode of valueNode.values) {
        if (isMissingVariable(itemNode, variables)) {
          if (isNonNullType(itemType)) {
            return;
          }
          coercedValues.push(null);
        } else {
          const itemValue = valueFromAST(itemNode, itemType, variables);
          if (void 0 === itemValue) {
            return;
          }
          coercedValues.push(itemValue);
        }
      }
      return coercedValues;
    }
    const coercedValue = valueFromAST(valueNode, itemType, variables);
    if (void 0 === coercedValue) {
      return;
    }
    return [ coercedValue ];
  }
  if (isInputObjectType(type)) {
    if (valueNode.kind !== e.OBJECT) {
      return;
    }
    const coercedObj = Object.create(null);
    const fieldNodes = keyMap(valueNode.fields, (field => field.name.value));
    for (const field of Object.values(type.getFields())) {
      const fieldNode = fieldNodes[field.name];
      if (!fieldNode || isMissingVariable(fieldNode.value, variables)) {
        if (void 0 !== field.defaultValue) {
          coercedObj[field.name] = field.defaultValue;
        } else if (isNonNullType(field.type)) {
          return;
        }
        continue;
      }
      const fieldValue = valueFromAST(fieldNode.value, field.type, variables);
      if (void 0 === fieldValue) {
        return;
      }
      coercedObj[field.name] = fieldValue;
    }
    return coercedObj;
  }
  if (isLeafType(type)) {
    let result;
    try {
      result = type.parseLiteral(valueNode, variables);
    } catch (_error) {
      return;
    }
    if (void 0 === result) {
      return;
    }
    return result;
  }
  invariant(!1);
}

function isMissingVariable(valueNode, variables) {
  return valueNode.kind === e.VARIABLE && (null == variables || void 0 === variables[valueNode.name.value]);
}

export { valueFromAST };
