import { invariant } from "../../jsutils/invariant.mjs";

import { OperationTypeNode } from "../../language/ast.mjs";

import { DirectiveLocation } from "../../language/directiveLocation.mjs";

import { specifiedDirectives } from "../../type/directives.mjs";

import { e, GraphQLError } from "../../shared.mjs";

function KnownDirectivesRule(context) {
  const locationsMap = Object.create(null);
  const schema = context.getSchema();
  const definedDirectives = schema ? schema.getDirectives() : specifiedDirectives;
  for (const directive of definedDirectives) {
    locationsMap[directive.name] = directive.locations;
  }
  const astDefinitions = context.getDocument().definitions;
  for (const def of astDefinitions) {
    if (def.kind === e.DIRECTIVE_DEFINITION) {
      locationsMap[def.name.value] = def.locations.map((name => name.value));
    }
  }
  return {
    Directive(node, _key, _parent, _path, ancestors) {
      const name = node.name.value;
      const locations = locationsMap[name];
      if (!locations) {
        context.reportError(new GraphQLError(`Unknown directive "@${name}".`, {
          nodes: node
        }));
        return;
      }
      const candidateLocation = function getDirectiveLocationForASTPath(ancestors) {
        const appliedTo = ancestors[ancestors.length - 1];
        "kind" in appliedTo || invariant(!1);
        switch (appliedTo.kind) {
         case e.OPERATION_DEFINITION:
          return function getDirectiveLocationForOperation(operation) {
            switch (operation) {
             case OperationTypeNode.QUERY:
              return DirectiveLocation.QUERY;

             case OperationTypeNode.MUTATION:
              return DirectiveLocation.MUTATION;

             case OperationTypeNode.SUBSCRIPTION:
              return DirectiveLocation.SUBSCRIPTION;
            }
          }(appliedTo.operation);

         case e.FIELD:
          return DirectiveLocation.FIELD;

         case e.FRAGMENT_SPREAD:
          return DirectiveLocation.FRAGMENT_SPREAD;

         case e.INLINE_FRAGMENT:
          return DirectiveLocation.INLINE_FRAGMENT;

         case e.FRAGMENT_DEFINITION:
          return DirectiveLocation.FRAGMENT_DEFINITION;

         case e.VARIABLE_DEFINITION:
          return DirectiveLocation.VARIABLE_DEFINITION;

         case e.SCHEMA_DEFINITION:
         case e.SCHEMA_EXTENSION:
          return DirectiveLocation.SCHEMA;

         case e.SCALAR_TYPE_DEFINITION:
         case e.SCALAR_TYPE_EXTENSION:
          return DirectiveLocation.SCALAR;

         case e.OBJECT_TYPE_DEFINITION:
         case e.OBJECT_TYPE_EXTENSION:
          return DirectiveLocation.OBJECT;

         case e.FIELD_DEFINITION:
          return DirectiveLocation.FIELD_DEFINITION;

         case e.INTERFACE_TYPE_DEFINITION:
         case e.INTERFACE_TYPE_EXTENSION:
          return DirectiveLocation.INTERFACE;

         case e.UNION_TYPE_DEFINITION:
         case e.UNION_TYPE_EXTENSION:
          return DirectiveLocation.UNION;

         case e.ENUM_TYPE_DEFINITION:
         case e.ENUM_TYPE_EXTENSION:
          return DirectiveLocation.ENUM;

         case e.ENUM_VALUE_DEFINITION:
          return DirectiveLocation.ENUM_VALUE;

         case e.INPUT_OBJECT_TYPE_DEFINITION:
         case e.INPUT_OBJECT_TYPE_EXTENSION:
          return DirectiveLocation.INPUT_OBJECT;

         case e.INPUT_VALUE_DEFINITION:
          {
            const parentNode = ancestors[ancestors.length - 3];
            "kind" in parentNode || invariant(!1);
            return parentNode.kind === e.INPUT_OBJECT_TYPE_DEFINITION ? DirectiveLocation.INPUT_FIELD_DEFINITION : DirectiveLocation.ARGUMENT_DEFINITION;
          }

         default:
          invariant(!1);
        }
      }(ancestors);
      if (candidateLocation && !locations.includes(candidateLocation)) {
        context.reportError(new GraphQLError(`Directive "@${name}" may not be used on ${candidateLocation}.`, {
          nodes: node
        }));
      }
    }
  };
}

export { KnownDirectivesRule };
