import { didYouMean } from "../../jsutils/didYouMean.mjs";

import { invariant } from "../../jsutils/invariant.mjs";

import { suggestionList } from "../../jsutils/suggestionList.mjs";

import { isTypeDefinitionNode } from "../../language/predicates.mjs";

import { isScalarType, isObjectType, isInterfaceType, isUnionType, isEnumType, isInputObjectType } from "../../type/definition.mjs";

import { GraphQLError, e } from "../../shared.mjs";

function PossibleTypeExtensionsRule(context) {
  const schema = context.getSchema();
  const definedTypes = Object.create(null);
  for (const def of context.getDocument().definitions) {
    if (isTypeDefinitionNode(def)) {
      definedTypes[def.name.value] = def;
    }
  }
  return {
    ScalarTypeExtension: checkExtension,
    ObjectTypeExtension: checkExtension,
    InterfaceTypeExtension: checkExtension,
    UnionTypeExtension: checkExtension,
    EnumTypeExtension: checkExtension,
    InputObjectTypeExtension: checkExtension
  };
  function checkExtension(node) {
    const typeName = node.name.value;
    const defNode = definedTypes[typeName];
    const existingType = null == schema ? void 0 : schema.getType(typeName);
    let expectedKind;
    if (defNode) {
      expectedKind = defKindToExtKind[defNode.kind];
    } else if (existingType) {
      expectedKind = function typeToExtKind(type) {
        if (isScalarType(type)) {
          return e.SCALAR_TYPE_EXTENSION;
        }
        if (isObjectType(type)) {
          return e.OBJECT_TYPE_EXTENSION;
        }
        if (isInterfaceType(type)) {
          return e.INTERFACE_TYPE_EXTENSION;
        }
        if (isUnionType(type)) {
          return e.UNION_TYPE_EXTENSION;
        }
        if (isEnumType(type)) {
          return e.ENUM_TYPE_EXTENSION;
        }
        if (isInputObjectType(type)) {
          return e.INPUT_OBJECT_TYPE_EXTENSION;
        }
        invariant(!1);
      }(existingType);
    }
    if (expectedKind) {
      if (expectedKind !== node.kind) {
        const kindStr = function extensionKindToTypeName(kind) {
          switch (kind) {
           case e.SCALAR_TYPE_EXTENSION:
            return "scalar";

           case e.OBJECT_TYPE_EXTENSION:
            return "object";

           case e.INTERFACE_TYPE_EXTENSION:
            return "interface";

           case e.UNION_TYPE_EXTENSION:
            return "union";

           case e.ENUM_TYPE_EXTENSION:
            return "enum";

           case e.INPUT_OBJECT_TYPE_EXTENSION:
            return "input object";

           default:
            invariant(!1);
          }
        }(node.kind);
        context.reportError(new GraphQLError(`Cannot extend non-${kindStr} type "${typeName}".`, {
          nodes: defNode ? [ defNode, node ] : node
        }));
      }
    } else {
      const allTypeNames = Object.keys({
        ...definedTypes,
        ...null == schema ? void 0 : schema.getTypeMap()
      });
      const suggestedTypes = suggestionList(typeName, allTypeNames);
      context.reportError(new GraphQLError(`Cannot extend type "${typeName}" because it is not defined.` + didYouMean(suggestedTypes), {
        nodes: node.name
      }));
    }
  }
}

const defKindToExtKind = {
  [e.SCALAR_TYPE_DEFINITION]: e.SCALAR_TYPE_EXTENSION,
  [e.OBJECT_TYPE_DEFINITION]: e.OBJECT_TYPE_EXTENSION,
  [e.INTERFACE_TYPE_DEFINITION]: e.INTERFACE_TYPE_EXTENSION,
  [e.UNION_TYPE_DEFINITION]: e.UNION_TYPE_EXTENSION,
  [e.ENUM_TYPE_DEFINITION]: e.ENUM_TYPE_EXTENSION,
  [e.INPUT_OBJECT_TYPE_DEFINITION]: e.INPUT_OBJECT_TYPE_EXTENSION
};

export { PossibleTypeExtensionsRule };
